/****************************************************************************
*  This file is part of the AVRIDE device driver
*
*  Copyright (C) 2002  Michael Fischer <mifi42@web.de>
*
*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
*  GNU General Public License for more details.
*
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
****************************************************************************
*  History:
*
*  14.12.02  mifi   First Version 
*  23.12.02  mifi   Add FileOpen, FileClose, FileError, FileSize,
*                   FileSeek and FileRead. But the FileSeek function
*                   does not work in the moment, later...
*  28.12.02  mifi   Now support FAT16 AND FAT32
****************************************************************************/
#define __FAT32_C__

#include <string.h>
#include <stddef.h>
#include <ctype.h>

#include <sys/heap.h>
#include <sys/print.h>
#include <sys/event.h>
#include <sys/thread.h>

#include "ide.h"
#include "fat32.h"
#include "typedefs.h"

/*==========================================================*/
/*  DEFINE: All Structures and Common Constants             */
/*==========================================================*/
#define FAT32_MAX_DRIVE                 IDE_MAX_SUPPORTED_DEVICE

//
// Some defines for the FAT structures
//
#define BPB_RsvdSecCnt                  32
#define BPB_NumFATs                     2
#define BPB_HiddSec                     63

#define FAT32_MEDIA                     0xf8

#define FAT32_OFFSET_FSINFO             1
#define FAT32_OFFSET_BACKUP_BOOT        6

#define FAT16_CLUSTER_EOF               0x0000FFFF
#define FAT16_CLUSTER_ERROR             0x0000FFF7
#define FAT16_CLUSTER_MASK              0x0000FFFF

#define FAT32_CLUSTER_EOF               0x0FFFFFFF
#define FAT32_CLUSTER_ERROR             0x0FFFFFF7
#define FAT32_CLUSTER_MASK              0x0FFFFFFF

#define FAT_SIGNATURE                   0xAA55

#define MBR_SIGNATURE                   FAT_SIGNATURE
#define MBR_FAT32                       0x0C

#define FSINFO_FIRSTSIGNATURE           0x41615252
#define FSINFO_FSINFOSIGNATURE          0x61417272
#define FSINFO_SIGNATURE                FAT_SIGNATURE

#define DIRECTORY_ATTRIBUTE_READ_ONLY   0x01
#define DIRECTORY_ATTRIBUTE_HIDDEN      0x02
#define DIRECTORY_ATTRIBUTE_SYSTEM_FILE 0x04
#define DIRECTORY_ATTRIBUTE_VOLUME_ID   0x08
#define DIRECTORY_ATTRIBUTE_DIRECTORY   0x10
#define DIRECTORY_ATTRIBUTE_ARCHIVE     0x20


#define FAT_NAME_LEN                    8
#define FAT_EXT_LEN                     3

//
//  DiskSize to SectorPerCluster table
//
typedef struct {
  DWORD DiskSize;
  BYTE  SecPerClusVal;
} DSKSZTOSECPERCLUS;

typedef struct _FAT32FileDataTime {
  unsigned Seconds : 5;
  unsigned Minute  : 6;
  unsigned Hour    : 5;
  unsigned Day     : 5;
  unsigned Month   : 4;
  unsigned Year    : 7;
} FAT32_FILEDATETIME, *PFAT32_FILEDATETIME;

typedef struct _FAT32DirectoryEntry {
  char               Name[FAT_NAME_LEN];
  char               Extension[FAT_EXT_LEN];
  BYTE               Attribute;
  BYTE               Reserved[8];
  WORD               HighCluster;
  FAT32_FILEDATETIME Date;
  WORD               LowCluster;
  DWORD              FileSize;
} FAT32_DIRECTORY_ENTRY;

typedef struct _FAT32DirectoryEntryLong {
  BYTE               Order;
  char               Name1[10];
  BYTE               Attribute;
  BYTE               Type;
  BYTE               Chksum;
  char               Name2[12];
  WORD               LowCluster;
  char               Name3[4];
} FAT32_DIRECTORY_ENTRY_LONG;

typedef struct _FAT32FileSystemInformation {
  DWORD FirstSignature;
  BYTE  Reserved1[480];
  DWORD FSInfoSignature;
  DWORD NumberOfFreeClusters;
  DWORD MostRecentlyAllocatedCluster;
  BYTE  Reserved2[12];
  BYTE  Reserved3[2];
  WORD  Signature;
} FAT32_FSINFO;

typedef struct _FAT32PartitionEntry {
  BYTE  BootInd;
  BYTE  FirstHead;
  BYTE  FirstSector;
  BYTE  FirstTrack;
  BYTE  FileSystem;
  BYTE  LastHead;
  BYTE  LastSector;
  BYTE  LastTrack;
  DWORD StartSectors;
  DWORD NumSectors;
} FAT32_PARTITION_ENTRY;

typedef struct _FAT32PartionTable {
 BYTE                  LoadInstruction[446];
 FAT32_PARTITION_ENTRY Partition[4];
 WORD                  Signature; /* AA55 */
} FAT32_PARTITION_TABLE;

typedef struct _bpbfat16 {
  BYTE  DrvNum;
  BYTE  Reserved1;
  BYTE  BootSig;
  DWORD VollID;
  BYTE  VolLab[11];      
  BYTE  FilSysType[8];
  BYTE  Reserved2[28];
} BPBFAT16;

typedef struct _bpbfat32 {
 DWORD FATSz32;                     // xxx
 WORD  ExtFlags;                    // 0
 WORD  FSVer;                       // must 0
 DWORD RootClus;                    // 
 WORD  FSInfo;                      // typically 1
 WORD  BkBootSec;                   // typically 6
 BYTE  Reserved[12];                // set all to zero
 BYTE  DrvNum;                      // must 0x80
 BYTE  Reserved1;                   // set all to zero
 BYTE  BootSig;                     // must 0x29 
 DWORD VollID;                      // xxx
 BYTE  VolLab[11];                  // "abcdefghijk"
 BYTE  FilSysType[8];               // "FAT32   "
} BPBFAT32;

typedef union _bpboffset36 {
  BPBFAT16 FAT16;
  BPBFAT32 FAT32;
} BPBOFFSET36;

typedef struct _FAT32BootRecord {
 BYTE        JumpBoot[3];           // 0xeb, 0x58, 0x90
 BYTE        OEMName[8];            // "MSWIN4.1"
 WORD        BytsPerSec;            // must 512
 BYTE        SecPerClus;            // 8 for 4K cluster
 WORD        RsvdSecCnt;            // typically 32 for FAT32
 BYTE        NumFATs;               // always 2
 WORD        RootEntCnt;            // must 0 for FAT32
 WORD        TotSec16;              // must 0 for FAT32
 BYTE        Media;                 // must 0xf8
             
 WORD        FATSz16;               // 0   for FAT32
 WORD        SecPerTrk;             // 63  for FAT32
 WORD        NumHeads;              // 255 for FAT32
 DWORD       HiddSec;               // 63  for FAT32
             
 DWORD       TotSec32;              // xxx

 BPBOFFSET36 Off36;

 BYTE        Reserved[420];
 WORD        Signature;             // must 0xAA55
} FAT32_BOOT_RECORD, *PFAT32_BOOT_RECORD;

typedef struct _fat_entry_table16 {
  WORD aEntry[256];
} FAT_ENTRY_TABLE16;

typedef struct _fat_entry_table32 {
  DWORD aEntry[128];
} FAT_ENTRY_TABLE32;

typedef union _fat_dir_table {
  FAT32_DIRECTORY_ENTRY      aShort[16];
  FAT32_DIRECTORY_ENTRY_LONG aLong[16];
} FAT_DIR_TABLE;

typedef struct _drive_info {
  BYTE   bIsFAT32;
  BYTE   bDevice;
  BYTE   bSectorsPerCluster;
  BYTE   bReserved;
  
  DWORD dwRootDirSectors;
  DWORD dwFirstRootDirSector;

  DWORD dwRootCluster;  
  DWORD dwFAT1StartSector;
  DWORD dwFAT2StartSector;
  DWORD dwCluster2StartSector;

  DWORD dwClusterSize;
} DRIVE_INFO;

typedef struct _fhandle {
  int         nFileID;
  DWORD      dwFileSize;
  DWORD      dwStartCluster;
  DWORD      dwReadCluster;
  DWORD      dwFilePointer;      /* total file pointer   */
  DWORD      dwClusterPointer;   /* cluster read pointer */

  int         nLastError;
  int         nEOF;

  DRIVE_INFO *pDrive;
} FHANDLE;

/*==========================================================*/
/*  DEFINE: Prototypes                                      */
/*==========================================================*/

/*==========================================================*/
/*  DEFINE: Definition of all local Data                    */
/*==========================================================*/
static BYTE      *pSectorBuffer;
static DRIVE_INFO sDriveInfo[FAT32_MAX_DRIVE];
static FHANDLE    aFileHandle[FAT32_MAX_FILES];

/*==========================================================*/
/*  DEFINE: Definition of all local Procedures              */
/*==========================================================*/
#define USE_SEMA    0

static volatile int FATSemaphore = 1;
static HANDLE hFATSemaphore;

void FATLock (void)
{
#if USE_SEMA
  while (FATSemaphore == 0) {
    NutThreadYield();
  }
  FATSemaphore--;
#else
  NutEventWait(&hFATSemaphore, 0);
#endif
}

void FATFree (void)
{
#if USE_SEMA
  FATSemaphore++;
#else
  NutEventPost(&hFATSemaphore);
#endif
}

void FATSemaInit(void)
{
#if USE_SEMA
  FATSemaphore = 1;
#else  
  NutEventPost(&hFATSemaphore);
#endif  
}

#if (FAT32_SUPPORT_FILEOPEN == 1)
/************************************************************/
/*  GetFirstSectorOfCluster                                 */
/************************************************************/
static DWORD GetFirstSectorOfCluster (DRIVE_INFO *pDrive, 
                                      DWORD     dwCluster)
{
  DWORD dwSector;

  dwSector  = (dwCluster-2) * pDrive->bSectorsPerCluster;
  dwSector += pDrive->dwCluster2StartSector;

  return(dwSector);
}

/************************************************************/
/*  GetNextCluster                                          */
/************************************************************/
static DWORD GetNextCluster (DRIVE_INFO *pDrive, 
                             DWORD      dwCluster)
{
  DWORD             dwNextCluster;
  DWORD             dwSector;
  DWORD             dwIndex;
  FAT_ENTRY_TABLE16 *pFatTable16;
  FAT_ENTRY_TABLE32 *pFatTable32;

  if (pDrive->bIsFAT32 == TRUE) {
    //
    //  (IDE_SECTOR_SIZE / sizeof(long)) == 128
    // 
    dwSector  = (dwCluster / 128) + pDrive->dwFAT1StartSector;
    dwIndex   = dwCluster  % 128;
  
    IDEReadSectors(pDrive->bDevice, pSectorBuffer, dwSector, 1);
    pFatTable32 = (FAT_ENTRY_TABLE32*)pSectorBuffer;
  
    dwNextCluster = (pFatTable32->aEntry[dwIndex] & FAT32_CLUSTER_MASK);
    if( (dwNextCluster == FAT32_CLUSTER_EOF)   ||
        (dwNextCluster == FAT32_CLUSTER_ERROR) ){
      dwNextCluster = 0;
    } 

  } else { /* FAT16 */
    //
    //  (IDE_SECTOR_SIZE / sizeof(word)) == 256
    // 
    dwSector  = (dwCluster / 256) + pDrive->dwFAT1StartSector;
    dwIndex   = dwCluster  % 256;
  
    IDEReadSectors(pDrive->bDevice, pSectorBuffer, dwSector, 1);
    pFatTable16 = (FAT_ENTRY_TABLE16*)pSectorBuffer;
  
    dwNextCluster = (pFatTable16->aEntry[dwIndex] & FAT16_CLUSTER_MASK);
    if( (dwNextCluster == FAT16_CLUSTER_EOF)   ||
        (dwNextCluster == FAT16_CLUSTER_ERROR) ){
      dwNextCluster = 0;
    } 

  } /* endif pDrive->bIsFAT32 */

  return (dwNextCluster);
}

/************************************************************/
/*  FindFile                                                */
/************************************************************/
static int FindFile (DRIVE_INFO            *pDrive, 
                     FAT32_DIRECTORY_ENTRY *pSearchEntry, 
                     DWORD                 dwDirCluster)
{
  int                    i,x;
  int                   nDirMaxSector;
  BYTE                  bFound;
  int                   nError;
  DWORD                 dwSector;
  FAT32_DIRECTORY_ENTRY *pDirEntry;
  FAT_DIR_TABLE         *pDirTable;

  nError = TRUE;
  bFound = FALSE;
  while( (bFound == FALSE) && (dwDirCluster != 0) ){
    dwSector      = GetFirstSectorOfCluster(pDrive, dwDirCluster);
    nDirMaxSector = (int)pDrive->bSectorsPerCluster;
    
    //
    // Test for special case dwDirCluster and FAT16
    //
    if( (dwDirCluster == 1 ) && (pDrive->bIsFAT32 == FALSE) ){
      dwSector      = pDrive->dwFirstRootDirSector;
      nDirMaxSector = (int)pDrive->dwRootDirSectors;
    }
    
    //
    // One cluster has SecPerCluster sectors
    //
    for(i=0; i<nDirMaxSector; i++) {
      IDEReadSectors(pDrive->bDevice, pSectorBuffer, dwSector+i, 1);
      pDirTable = (FAT_DIR_TABLE*)pSectorBuffer;

      //
      // And one sector has 16 entries
      //
      // IDE_SECTOR_SIZE / sizeof(FAT32_DIRECTORY_ENTRY) = 16
      //
      for(x=0; x<16; x++) {
        //
        // Check for valid entry
        //
        pDirEntry = (FAT32_DIRECTORY_ENTRY*)pDirTable->aShort[x].Name;
        if( (pDirEntry->Name[0] != 0xE5) && (pDirEntry->Name[0] != 0x00) ){
          if (memcmp(pDirEntry, pSearchEntry, 11) == 0) {
            //
            // Check for the correct attribute, this is done with
            // the '&' and not with the memcmp.
            // With the '&' it is possible to find a hidden archive too :-)
            //
            if( (pDirEntry->Attribute & pSearchEntry->Attribute) == pSearchEntry->Attribute) {
              bFound                    = TRUE;
              pSearchEntry->HighCluster = pDirEntry->HighCluster;
              pSearchEntry->LowCluster  = pDirEntry->LowCluster;
              pSearchEntry->FileSize    = pDirEntry->FileSize;
              nError                    = FALSE; 
              break;
            } /* endif Attribute */
          } /* endif test Name+Ext */
        } /* endif Name[0] != 0xe5, 0x00 */
      } /* endfor x<16 */

      if(bFound == TRUE) {
        break;
      }
    }     

    if (bFound == FALSE) {
      //
      // No file found in this cluster, get the next one
      //
      dwDirCluster = GetNextCluster(pDrive, dwDirCluster);
    } 
  } /* endwhile */

  return (nError);
}
#endif

/*==========================================================*/
/*  DEFINE: All code exported                               */
/*==========================================================*/
/************************************************************/
/*  FAT32Init                                               */
/************************************************************/
int FAT32Init (void)
{
  BYTE                   i;
  int                   nError;
  DWORD                 dwSector;
  DWORD                 dwFATSz;
  DWORD                 dwRootDirSectors;
  FAT32_PARTITION_TABLE *pPartitionTable;
  FAT32_BOOT_RECORD     *pBootRecord; 

  FATSemaInit();

  pSectorBuffer = (BYTE *)NutHeapAlloc(IDE_SECTOR_SIZE);
  if (pSectorBuffer != NULL) {

    //
    // Clear all FileHandle
    //
    for(i=0; i<FAT32_MAX_FILES; i++) {
      memset(&aFileHandle[i], 0x00, sizeof(FHANDLE));
    }
    
    for(i=0; i<FAT32_MAX_DRIVE; i++) {
      memset((BYTE *)&sDriveInfo[i], 0x00, sizeof(DRIVE_INFO));
    
      sDriveInfo[i].bDevice = i;
    
      //
      // Try to find a PartitionTable
      // 
      nError = IDEReadSectors(i, pSectorBuffer, 0, 1);
      if (nError == IDE_OK) {
        pPartitionTable = (FAT32_PARTITION_TABLE*)pSectorBuffer;   
    
        if (pPartitionTable->Signature == FAT_SIGNATURE) {
            if (pPartitionTable->Partition[0].NumSectors) {
            
            //
            // We found a PartitionTable, read BootRecord
            // 
            dwSector = pPartitionTable->Partition[0].StartSectors;
            IDEReadSectors(i, pSectorBuffer, dwSector, 1);
            pBootRecord = (FAT32_BOOT_RECORD*)pSectorBuffer;
    
            //
            // Test valid BootRecord
            //
            if (pBootRecord->Signature = FAT_SIGNATURE) {
              sDriveInfo[i].bSectorsPerCluster = pBootRecord->SecPerClus;
      
              if (pBootRecord->FATSz16 != 0) {
                dwFATSz                     = pBootRecord->FATSz16;
                sDriveInfo[i].bIsFAT32      = FALSE;
                sDriveInfo[i].dwRootCluster = 1; /* special value, see */
                                                 /* FindFile           */
              } else {
                dwFATSz                     = pBootRecord->Off36.FAT32.FATSz32;
                sDriveInfo[i].bIsFAT32      = TRUE;
                sDriveInfo[i].dwRootCluster = pBootRecord->Off36.FAT32.RootClus;
              } 
    
              dwRootDirSectors = ((pBootRecord->RootEntCnt * 32) +
                                  (pBootRecord->BytsPerSec - 1)) / pBootRecord->BytsPerSec;
    
              sDriveInfo[i].dwFAT1StartSector = 
                pBootRecord->HiddSec + pBootRecord->RsvdSecCnt;
              
              sDriveInfo[i].dwFAT2StartSector = 
                sDriveInfo[i].dwFAT1StartSector + dwFATSz;
              
              sDriveInfo[i].dwCluster2StartSector =
                pBootRecord->HiddSec    + 
                pBootRecord->RsvdSecCnt +
                (pBootRecord->NumFATs * dwFATSz) + dwRootDirSectors;                                                 
    
              sDriveInfo[i].dwClusterSize = 
                pBootRecord->SecPerClus * IDE_SECTOR_SIZE;
    
              sDriveInfo[i].dwRootDirSectors     = dwRootDirSectors;
              sDriveInfo[i].dwFirstRootDirSector = sDriveInfo[i].dwFAT2StartSector + dwFATSz;           
                  
            } /* endif pBootRecord->Signature */
          } /* endif pPartitionTable->Partition[0].NumSectors */
        } /* endif pPartitionTable->Signature == FAT_SIGNATURE */
      } /* endif IDEReadSectors == IDE_OK */
    } /* endfor i<FAT32_MAX_DRIVE */
  } /* endif pSectorBuffer != NULL */

  return (nError);
}

#if (FAT32_SUPPORT_FORMAT == 1)
/************************************************************/
/*  FAT32Format                                             */
/************************************************************/
int FAT32Format (BYTE bDevice)
{
  int nError = FAT32_ERROR;

  return (nError);
}
#endif

#if (FAT32_SUPPORT_FILEOPEN == 1)
/************************************************************/
/*  FAT32FileOpen                                           */
/*                                                          */
/*  Opens an existing file for reading.                     */
/*                                                          */
/*  Parameters: pName points to a string that specifies the */
/*              name of the file to open. The name must     */
/*              exactly match the full pathname of the file.*/
/*                                                          */
/*  Returns:    A pointer to a FILE structure that can be   */
/*              used to read the file and retrieve          */
/*              information about the file.                 */
/************************************************************/
FILE *FAT32FileOpen (char *pName)
{
  int                     i;
  int                    nEndWhile;
  DWORD                 dwCluster;
  DWORD                 dwSector;
  FHANDLE               *hReturn;
  FHANDLE               *hFile;
  DRIVE_INFO            *pDrive;
  FAT32_DIRECTORY_ENTRY sDirEntry;

  FATLock();

  hReturn = NULL;
  hFile   = NULL;

  if (pName[0] != 0) {
    //
    // @@MF FIX ME later, we support only IDE_DRIVE_C
    //
    pDrive = &sDriveInfo[IDE_DRIVE_C];
    
    //
    // Find a free file handle
    //
    for(i=0; i<FAT32_MAX_FILES; i++) {
      if (aFileHandle[i].dwFileSize == 0) {
        //
        // If the FileSize is ZERO, this handle is free
        //
        hFile = (FHANDLE *)&aFileHandle[i];
        break;
      }
    }
    
    if( (pDrive->dwFAT1StartSector) && (hFile != NULL) && (*pName != '.') ){
      memset(&sDirEntry, 0x00, sizeof(FAT32_DIRECTORY_ENTRY));
    
      //
      // Start by the ROOT dir...
      //
      dwCluster = pDrive->dwRootCluster;
    
      nEndWhile = FALSE;
      while (nEndWhile == FALSE) { /* master loop */
    
        memset(sDirEntry.Name,      0x20, FAT_NAME_LEN);
        memset(sDirEntry.Extension, 0x20, FAT_EXT_LEN);
    
        //
        // Get Name
        //
        i = 0;
        while( (*pName != '/') && (*pName != '\\') && 
               (*pName != '.') && (*pName != 0)    ){
    
          if (i == FAT_NAME_LEN) {
            nEndWhile = TRUE;
            break;
          }
          
          sDirEntry.Name[i] = toupper(*pName);
    
          i++;
          pName++;
        } /* endwhile Name */
    
        if( (nEndWhile == FALSE) && (*pName == '.') ){
          pName++; /* jump over '.' */
    
          //
          // Get Extension
          //
          i = 0;
          while( (*pName != '/') && (*pName != '\\') && (*pName != 0) ){
          
            if (i >= FAT_EXT_LEN) {

              //
              // This is an ERROR, but in the moment we does not 
              // support extensions more than 3 chars.
              //       
#if 0             
              nEndWhile = TRUE;
              break;
#endif            
            } else {          
              sDirEntry.Extension[i] = toupper(*pName);
            }  
          
            i++;
            pName++;
          } /* endwhile Extension */
        }
    
        if (nEndWhile == TRUE) {
          /* break out of the master loop */
          break;
        }
    
        switch (*pName) {
          //
          // The file is an ARCHIVE
          //
          case 0: { 
            nEndWhile = TRUE;
            sDirEntry.Attribute = DIRECTORY_ATTRIBUTE_ARCHIVE;
            if (FindFile(pDrive, &sDirEntry, dwCluster) == FALSE) {
              if (sDirEntry.FileSize != 0) {
                //
                // NO Error
                //
                dwCluster = sDirEntry.HighCluster;
                dwCluster = (dwCluster << 16) | (DWORD)sDirEntry.LowCluster;
                dwSector  = GetFirstSectorOfCluster(pDrive, dwCluster);
              
                hFile->dwFileSize       = sDirEntry.FileSize;
                hFile->dwStartCluster   = dwCluster;
                hFile->dwReadCluster    = dwCluster;
                hFile->dwFilePointer    = 0;
                hFile->dwClusterPointer = 0;
                hFile->pDrive           = pDrive;
                hFile->nLastError       = FAT32_OK;
                hFile->nEOF             = FALSE;
              
                hReturn = hFile;
              } /* endif sDirEntry.FileSize != 0 */
            } /* endif FindFile */
            break;
          } /* endcase 0 */
    
          //
          // The file is a DIRECTORY
          //
          case '/':
          case '\\': {
            pName++; /* jump over the char */
    
            sDirEntry.Attribute = DIRECTORY_ATTRIBUTE_DIRECTORY;
            if (FindFile(pDrive, &sDirEntry, dwCluster) == FALSE) {
              dwCluster = sDirEntry.HighCluster;
              dwCluster = (dwCluster << 16) | (DWORD)sDirEntry.LowCluster;
    
              //
              // The new Cluster is the Cluster of the directory
              //
    
            } else {
              nEndWhile = TRUE;
            }
            break;
          }
    
          default: {
            nEndWhile = TRUE;
            break;
          }
        } /* end switch */
    
      } /* end while */
    
    } /* endif pDrive->dwFAT1StartSector */

  } /* endif pName[0] != 0 */

  FATFree();

  return (hReturn);
}
#endif

#if (FAT32_SUPPORT_FILECLOSE == 1)
/************************************************************/
/*  FAT32FileClose                                          */
/*                                                          */
/*  Close a previously opened file.                         */
/*                                                          */
/*  Parameters: pFile Identifies the file to close. This    */
/*              pointer must have been created by calling   */
/*              FAT32FileOpen().                            */
/*                                                          */
/*  Returns:    0 if the function is successful,            */
/*              -1 otherwise.                               */
/************************************************************/
int FAT32FileClose (FILE *pFile)
{
  int      nError;
  FHANDLE *hFile;

  FATLock();

  hFile  = (FHANDLE*)pFile;
  nError = FAT32_ERROR;

  if (hFile != NULL) {
    hFile->dwFileSize = 0;
    nError = FAT32_OK;
  }

  FATFree();

  return (nError);
}
#endif

#if (FAT32_SUPPORT_FILEERROR == 1)
/************************************************************/
/*  FAT32FileError                                          */
/*                                                          */
/*  Test for an error on a file.                            */
/*                                                          */
/*  Parameters: pFile Identifies the file to test.          */
/*                                                          */
/*  Returns:    0, if no error has occurred.                */
/*              Otherwise, it returns a nonzero value.      */
/************************************************************/
int FAT32FileError (FILE *pFile)
{
  int      nError = FAT32_OK;
  FHANDLE *hFile;

  FATLock();

  hFile = (FHANDLE*)pFile;

  if (hFile != NULL) {
    nError = hFile->nLastError;
  } else {
    nError = FAT32_ERROR_HANDLE;
  }

  FATFree();

  return (nError);
}
#endif

#if (FAT32_SUPPORT_FILESIZE == 1)
/************************************************************/
/*  FAT32FileSize                                           */
/*                                                          */
/*  Retrieve the size of a file.                            */
/*                                                          */
/*  Parameters: pFile Identifies the file to query.         */
/*              This pointer must have been created by      */
/*              calling FAT32FileOpen().                    */
/*                                                          */
/*  Returns:    The number of bytes in this file or         */
/*              -1 if an error occured                      */
/************************************************************/
long FAT32FileSize (FILE *pFile)
{
  long     lSize = -1;
  FHANDLE *hFile;

  FATLock();

  hFile = (FHANDLE*)pFile;

  if (hFile != NULL) {
    lSize = hFile->dwFileSize;
  }

  FATFree();

  return (lSize);
}
#endif


#if (FAT32_SUPPORT_FILESEEK == 1)
/************************************************************/
/*  FAT32FileSeek                                           */
/*                                                          */
/*  Move the file pointer to a new position.                */
/*  It points to the next byte to be read from a file.      */
/*  The file pointer is automatically incremented for       */
/*  each byte read. When the file is opened, it is at       */
/*  position 0, the beginning of the file.                  */
/*                                                          */
/*  Parameters: pFile Identifies the file to seek.          */
/*              This pointer must have been created by      */
/*              calling FAT32FileOpen().                    */
/*                                                          */
/*              lPos Specifies the new absolute position    */
/*              of the file pointer.                        */
/*                                                          */
/*  Returns:    0 if the function is successful,            */
/*              -1 otherwise.                               */
/************************************************************/
int FAT32FileSeek (FILE *pFile, long lPos)
{
  int      nError = FAT32_OK;
  FHANDLE *hFile;

  FATLock();

  hFile = (FHANDLE*)pFile;

  //
  // We must do some work here...
  //
  nError = FAT32_ERROR;

  FATFree();

  return (nError);
}
#endif

#if (FAT32_SUPPORT_FILEREAD == 1)
/************************************************************/
/*  FAT32FileRead                                           */
/*                                                          */
/*  Read data from a file.                                  */
/*                                                          */
/*  Parameters: pFile Identifies the file to read from.     */
/*              This pointer must have been created by      */
/*              calling FAT32FileOpen().                    */
/*                                                          */
/*              pData Points to the buffer that receives    */
/*              the data.                                   */
/*                                                          */
/*              nSize Specifies the number of bytes to      */
/*              read from the file.                         */
/*                                                          */
/*  Returns:    The number of bytes read from the file or   */
/*               -1 if an error occured.                    */
/************************************************************/
int FAT32FileRead (FILE *pFile, void *pData, int nSize)
{
  int         nError;
  int         nBytesRead = 0;
  int         nBytesToRead;
  FHANDLE    *hFile;
  DRIVE_INFO *pDrive;
  BYTE       *pByte;
  DWORD      dwReadSector;
  DWORD      dwSector;
  int         nSectorCount;
  int         nSectorOffset;

  FATLock();

  hFile = (FHANDLE*)pFile;

  if( (hFile != NULL) && (nSize != 0) ){
    if (hFile->dwFilePointer < hFile->dwFileSize) {

      if ((hFile->dwFilePointer + nSize) > hFile->dwFileSize) {
        nSize = (int)(hFile->dwFileSize - hFile->dwFilePointer);
      }
      
      pDrive = (DRIVE_INFO*)hFile->pDrive;
      pByte  = (BYTE*)pData;

      nBytesRead = nSize;
      
      while (nSize) {
        dwSector      = GetFirstSectorOfCluster(pDrive, hFile->dwReadCluster);
        nSectorCount  = hFile->dwClusterPointer / IDE_SECTOR_SIZE;
        nSectorOffset = hFile->dwClusterPointer % IDE_SECTOR_SIZE;
        
        //
        // (Sector + SectorCount) is the sector to read
        // SectorOffset is the start position in the sector itself
        //
        dwReadSector = dwSector + nSectorCount;
        nError = IDEReadSectors(pDrive->bDevice, 
                                pSectorBuffer, dwReadSector, 1);
      
        if (nError == IDE_OK) { 
          //
          // Find the size we can read from ONE sector
          //
          if (nSize > IDE_SECTOR_SIZE) {
            nBytesToRead = IDE_SECTOR_SIZE;
          } else {
            nBytesToRead = nSize;
          }
          
          //
          // Test inside a sector
          //
          if ((nSectorOffset + nBytesToRead) > IDE_SECTOR_SIZE) {
            nBytesToRead = IDE_SECTOR_SIZE - nSectorOffset;
          }
          
          memcpy(pByte, &pSectorBuffer[nSectorOffset], nBytesToRead);
          pByte += nBytesToRead;
          
          hFile->dwFilePointer    += nBytesToRead;
          hFile->dwClusterPointer += nBytesToRead;
          
          //
          // Check for EOF
          if (hFile->dwFilePointer >= hFile->dwFileSize) {
            hFile->nEOF = TRUE;
          }
          
          if (hFile->dwClusterPointer >= pDrive->dwClusterSize) {
            //
            // We must switch to the next cluster
            //
            hFile->dwReadCluster    = GetNextCluster(pDrive, hFile->dwReadCluster);
            hFile->dwClusterPointer = 0;
          }                     
                             
          nSize -= nBytesToRead;

        } else { /* IDEReadSectors Error */

          nBytesRead        = 0;
          hFile->nLastError = FAT32_ERROR_IDE;
          break;
        } /* endif nError == IDE_OK */

      } /* endwhile */

    } else { /* reached the EOF */
      hFile->nLastError = FAT32_ERROR_EOF;
    } /* endif hFile->dwFilePointer < hFile->dwFileSize */

  } /* endif (hFile != NULL) && (nSize != 0) */

  FATFree();

  return (nBytesRead);
}
#endif

FILE *NutFileOpen(char *name)
{
  FILE    *hFile;
//  FHANDLE *pFile;
  
  hFile = FAT32FileOpen(name);
  if (hFile == NULL) {
//    NutPrintFormat(0, "FileOpen: \"%s\"-Error\r\n", name);
  } else {
//    pFile = (FHANDLE *)hFile;
//    NutPrintFormat(0, "FileOpen[%d]: \"%s\"-Ok\r\n", pFile->nFileID, name); 
  }

  return (hFile);
}

int NutFileClose (FILE *romf)
{
  int     nError;
//  FHANDLE *pFile = (FHANDLE *)romf;

//  NutPrintFormat(0, "FileClose[%d]\r\n", pFile->nFileID);

  nError = FAT32FileClose(romf);

  return (nError);
}

int NutFileRead(FILE *romf, void *data, int size)
{
  int      nError;
//  FHANDLE *pFile = (FHANDLE *)romf;

  nError = FAT32FileRead(romf, data, size);
//  NutPrintFormat(0, "FileRead[%d] %d\r\n", pFile->nFileID, size);

  return (nError);
}

int NutFileSize(FILE *romf)
{
  int      nFileSize;
//  FHANDLE *pFile = (FHANDLE *)romf;

  nFileSize = FAT32FileSize(romf);
//  NutPrintFormat(0, "FileSize[%d] %d\r\n", pFile->nFileID, nFileSize);

  return (nFileSize);
}


