/****************************************************************************
*  This file is part of the AVRIDE device driver
*
*  Copyright (C) 2002  Michael Fischer <mifi42@web.de>
*
*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
*  GNU General Public License for more details.
*
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
****************************************************************************
*  History:
*
*  14.12.02  mifi   First Version
*  29.12.02  mifi   Support CF-Card too 
****************************************************************************/
#define __IDE_C__

#include <stddef.h>
#include <macros.h>

#include <sys/timer.h>
#include <sys/event.h>
#include <dev/irqreg.h>
#include <sys/atom.h>

#include "ide.h"
#include "idep.h"
#include "typedefs.h"

/*==========================================================*/
/*  DEFINE: All Structures and Common Constants             */
/*==========================================================*/
#define USE_COMPACT_FLASH       TRUE
//#define USE_COMPACT_FLASH       FALSE

#define INITTIMEOUT             10000
#define DISKTIMEOUT             5000
#define CONTROLLERTIMEOUT       200

#define IDELock()   
#define IDEFree()   

#define IDEIn(x)                pIDE[x]
#define IDEOut(x,y)             pIDE[x] = y

//
// Drive Flags
//
#define IDE_SUPPORT_LBA         0x0001
#define IDE_SUPPORT_LBA48       0x0002
#define IDE_READY               0x8000

typedef struct _drive {
  //
  // Interface values
  //
  WORD   wFlags;
  BYTE   bCompactFlash;
  BYTE   bDevice; 

#if (IDE_SUPPORT_CHS == 1) 
  //
  // CHS values
  //
  WORD   wCylinders;
  WORD   wHeads;
  WORD   wSectorsPerTrack;
  DWORD dwOneSide;
#endif

  //
  // LBA value
  //
  DWORD dwTotalSectors;

} DRIVE, *LPDRIVE;

/*==========================================================*/
/*  DEFINE: Prototypes                                      */
/*==========================================================*/

/*==========================================================*/
/*  DEFINE: Definition of all local Data                    */
/*==========================================================*/
static HANDLE         hIDEEvent; 
static volatile BYTE *pIDE;
static volatile BYTE gbIntStatus = 0;

static DRIVE          sDrive[IDE_MAX_SUPPORTED_DEVICE];     
/*==========================================================*/
/*  DEFINE: Definition of all local Procedures              */
/*==========================================================*/
/************************************************************/
/*  HardwareReset                                           */
/************************************************************/
static void HardwareReset (DRIVE *pDrive)
{
  if (pDrive->bCompactFlash == TRUE) {
    //
    // Set Reset
    //
    PORTD |= 0x20;
    NOP();

    NutSleep(1000);

    //
    // Remove Reset
    //
    PORTD &= ~0x20;
    NOP();

    NutSleep(1000);    
  } else { /* HD */
    //
    // Set RESET
    //
    PORTD &= ~0x20;
    NOP();
  
    NutSleep(1000);
  
    //
    // Remove RESET
    //
    PORTD |= 0x20; 
    NOP();
  
    //
    // It can takes up to 10 seconds (2.5-Inch drive) 
    // from reset high to NOT busy.
    // Take a look in the drive manual "Reset timings"
    //
    //NutSleep(10000);
    NutSleep(4000);
  }
}

/************************************************************/
/*  IDEInterrupt                                            */
/************************************************************/
static void IDEInterrupt (void *p)
{
  p = p;
  
  gbIntStatus = IDEIn(STATUS_REG);
  
  NutEventPostAsync(&hIDEEvent);
}

/************************************************************/
/*  ClearEvent                                              */
/************************************************************/
static void ClearEvent (void)
{
  *((int *)hIDEEvent) = 0;
}

/************************************************************/
/*  WaitForInterrupt                                        */
/*                                                          */
/*  Return: IDE_OK, IDE_ERROR                               */
/************************************************************/
static int WaitForInterrupt (DWORD dwTimeout)
{
  int nError = IDE_OK;
  int nTimeout;

  nTimeout = NutEventWait(&hIDEEvent, dwTimeout);
  if (nTimeout == -1) {
    nError = IDE_ERROR;
  }

  return (nError);
}  

#if (IDE_SUPPORT_WRITE == 1)
/************************************************************/
/*  WaitDRQ                                                 */
/*                                                          */
/*  Return: IDE_OK, IDE_ERROR                               */
/************************************************************/
static int WaitDRQ (DWORD dwTimeout)
{
  int  nError = IDE_OK;
  BYTE bStatus;
 
  bStatus = IDEIn(STATUS_REG);
  if ((bStatus & (STATUS_BUSY | STATUS_DATA_REQUEST)) != STATUS_DATA_REQUEST) {

    dwTimeout  = (DWORD)(((dwTimeout*10UL) / 625UL) + 1UL);
    dwTimeout += NutGetTickCount();

    bStatus = IDEIn(STATUS_REG);
    while ((bStatus & (STATUS_BUSY | STATUS_DATA_REQUEST)) != STATUS_DATA_REQUEST) {
      if (NutGetTickCount() > dwTimeout) {
        nError = IDE_ERROR;
        break;
      }
      bStatus = IDEIn(STATUS_REG);
    }

  }

  return (nError);
}
#endif

/************************************************************/
/*  WaitNotBusy  (see ATAPI-4 r17 p223 9.6)                 */
/*                                                          */
/*  Return: IDE_OK, IDE_ERROR                               */
/************************************************************/
static int WaitNotBusy (DWORD dwTimeout)
{
  int  nError = IDE_OK;
  BYTE bStatus;

  bStatus = IDEIn(STATUS_REG);
  if (bStatus & (STATUS_BUSY | STATUS_DATA_REQUEST) ) {

    dwTimeout  = ((dwTimeout*10) / 625) + 1;
    dwTimeout += NutGetTickCount();

    bStatus = IDEIn(STATUS_REG);
    while ( bStatus & (STATUS_BUSY | STATUS_DATA_REQUEST) ) {
      if (NutGetTickCount() > dwTimeout) {
        nError = IDE_BUSY;
        break;
      }
      bStatus = IDEIn(STATUS_REG);
    }

  }

  return (nError);
}

/************************************************************/
/*  SelectDevice  (see ATAPI-4 r17 p223 9.6)                */
/*                                                          */
/*  Return: IDE_OK, IDE_ERROR                               */
/************************************************************/
static int SelectDevice (BYTE bDevice)
{
  int  nError;
  int  nTimeout;
  BYTE bStatus;
  BYTE bDummy;

  nError = IDE_OK;

  if (WaitNotBusy(CONTROLLERTIMEOUT) == IDE_OK) {

    IDEOut(DISK_HEAD_REG, 0xA0 + (bDevice << 4) );

    //
    // Wait at least 400ns, with 14.7456Mhz we need
    // 8 nop's for 500ns
    //
    COMPRESS_DISABLE;
    NOP();
    NOP();
    NOP();
    NOP();
    NOP();
    NOP();
    NOP();
    NOP();
    COMPRESS_REENABLE;

    nError = WaitNotBusy(CONTROLLERTIMEOUT);
  } else {
    nError = IDE_BUSY;
  }

  if (nError != IDE_OK) {
    //
    // Test for special case, data available.
    // If data available, read the data!!
    //
    bStatus = IDEIn(STATUS_REG);
    if (bStatus & STATUS_DATA_REQUEST) {
     
      nTimeout = 512; 
      while( (bStatus & STATUS_DATA_REQUEST) && (nTimeout) ){
        bDummy  = IDEIn(DATA_READ_REG_LOW);
        bDummy  = IDEIn(DATA_READ_REG_LOW);
        bStatus = IDEIn(STATUS_REG);
        nTimeout--;
      }
    }
  }

  return (nError);
}

/************************************************************/
/*  GetDeviceInfo                                           */
/*                                                          */
/*  PIO data in command protocol (see ATAPI-4 r17 p224 9.7) */
/*                                                          */
/*  Return: IDE_OK, IDE_ERROR                               */
/************************************************************/
static BYTE GetDeviceInfo (LPDRIVE pDrive, BYTE *pSectorBuffer)
{
  WORD           i;
  int            nError;
  BYTE           bStatus;
  BYTE           bErrorReg;
  BYTE           bDevice = pDrive->bDevice;
  IDEDEVICEINFO *pInfo;

  nError = IDE_ERROR;

  if (SelectDevice(bDevice) == IDE_OK) {

    ClearEvent();

    IDEOut(FEATURE_REG,       0);
    IDEOut(SECTOR_COUNT_REG,  1);
    IDEOut(SECTOR_REG,        0);
    IDEOut(CYLINDER_LOW_REG,  0);
    IDEOut(CYLINDER_HIGH_REG, 0);    

    IDEOut(COMMAND_REG, COMMAND_DEVICE_INFO);
    if (WaitForInterrupt(DISKTIMEOUT) == IDE_OK) {

      if (gbIntStatus & STATUS_DATA_REQUEST) {

        for(i=0; i<IDE_SECTOR_SIZE; i=i+2) {

          pSectorBuffer[i] = pIDE[DATA_READ_REG_LOW];
          
          if (pDrive->bCompactFlash) {    
            pSectorBuffer[i+1] = pIDE[DATA_READ_REG_LOW];
          } else {        
            pSectorBuffer[i+1] = pIDE[DATA_READ_REG_HIGH];
          }
          
        }

        //
        // Status register is read to clear
        // pending interrupts.
        //
        bStatus = IDEIn(STATUS_REG); 

        pInfo = (IDEDEVICEINFO *)pSectorBuffer;

#if (IDE_SUPPORT_CHS == 1) 
        pDrive->wCylinders       = (WORD)pInfo->Cylinders;
        pDrive->wHeads           = (BYTE)pInfo->Heads;
        pDrive->wSectorsPerTrack = (BYTE)pInfo->SectorsPerTrack;
        pDrive->dwOneSide        = pDrive->wCylinders * 
                                   pDrive->wSectorsPerTrack;
#endif

        if (pInfo->LBASectors > 0) {
          pDrive->wFlags          = IDE_READY | IDE_SUPPORT_LBA;
          pDrive->dwTotalSectors  = pInfo->LBASectors;
        } else {

#if (IDE_SUPPORT_CHS == 1) 
          pDrive->wFlags          = IDE_READY;
          pDrive->dwTotalSectors  = pInfo->Cylinders;
          pDrive->dwTotalSectors *= pInfo->Heads;
          pDrive->dwTotalSectors *= pInfo->SectorsPerTrack;
#endif

        }

        nError = IDE_OK;
      }

      if (gbIntStatus & STATUS_ERROR) {
        bErrorReg = IDEIn(ERROR_REG);
      }
    }
    
  }

  return (nError);
}

/************************************************************/
/*  DeviceDiag                                              */
/*                                                          */
/*  NON-data command protocol (see ATAPI-4 r17 p231 9.9)    */
/*                                                          */
/*  Return: IDE_OK, IDE_ERROR                               */
/************************************************************/
static int DeviceDiag (LPDRIVE pDrive)
{
  int  nError = IDE_ERROR;
  BYTE bResult;
  BYTE bDevice = pDrive->bDevice;

  if (SelectDevice(0) == IDE_OK) { 

    ClearEvent();

    IDEOut(FEATURE_REG,       0);
    IDEOut(SECTOR_COUNT_REG,  1);
    IDEOut(SECTOR_REG,        0);
    IDEOut(CYLINDER_LOW_REG,  0);
    IDEOut(CYLINDER_HIGH_REG, 0); 

    IDEOut(COMMAND_REG, COMMAND_DIAG);
    if (WaitForInterrupt(INITTIMEOUT) == IDE_OK) {
      bResult = IDEIn(ERROR_REG);

      if (bResult == 0x01) {  // see ATAPI-4 r17 p72
        nError = IDE_OK;
      }

    }
  }

  return (nError);
}

/************************************************************/
/*  ReadSectors                                             */
/*                                                          */
/*  PIO data in command protocol, see ATAPI-4 r17 p224 9.7  */
/*                                                          */
/*  Return: IDE_OK, IDE_ERROR                               */
/************************************************************/
static int ReadSectors (BYTE   bDevice,      BYTE *pData, 
                        DWORD dwStartSector, WORD  wSectorCount)
{
  WORD    i, x;
  int     nError;
  BYTE    bValue;
  BYTE    bErrorReg;
  BYTE    bStatus;
  LPDRIVE pDrive;
          
#if (IDE_SUPPORT_CHS == 1)
  WORD    wCHSSector;
  WORD    wCHSHead;
  WORD    wCHSCylinder;
#endif

  pDrive = &sDrive[bDevice];
  nError = SelectDevice(bDevice);

#if (IDE_SUPPORT_CHS == 0)
  //
  // NO CHS support
  //
  if ((pDrive->wFlags & IDE_SUPPORT_LBA) == 0) {
    //
    // If we have a CHS device, this is an ERROR
    //
    nError = IDE_ERROR;
  }
#endif

  if (nError == IDE_OK) {

    ClearEvent();

    if (pDrive->wFlags & IDE_SUPPORT_LBA) {
      //
      // LBA
      //
      bValue = (BYTE) ((bDevice << 4) | 0xE0 | (dwStartSector >> 24));
      IDEOut(DISK_HEAD_REG, bValue);
      IDEOut(LBA_16_23, (BYTE)(dwStartSector >> 16));
      IDEOut(LBA_8_15,  (BYTE)(dwStartSector >> 8));
      IDEOut(LBA_0_7,   (BYTE)(dwStartSector));

      IDEOut(SECTOR_COUNT_REG, (BYTE)(wSectorCount & 0xff));
      IDEOut(COMMAND_REG, COMMAND_READ_SECTORS);

#if (IDE_SUPPORT_CHS == 1)
    } else {
      //
      // CHS
      //
      wCHSSector    = (WORD) (dwStartSector % pDrive->wSectorsPerTrack) + 1;
      wCHSHead      = (WORD) (dwStartSector / pDrive->dwOneSide);
      dwStartSector = dwStartSector % pDrive->dwOneSide;
      wCHSCylinder  = (WORD) (dwStartSector / pDrive->wSectorsPerTrack);

      wCHSHead     |= (bDevice << 4) | 0xA0;

      IDEOut(DISK_HEAD_REG,     (BYTE)wCHSHead);
      IDEOut(SECTOR_REG,        (BYTE)wCHSSector);
      IDEOut(CYLINDER_LOW_REG,  (BYTE)(wCHSCylinder & 0x00FF));
      IDEOut(CYLINDER_HIGH_REG, (BYTE)(wCHSCylinder >> 8));

      IDEOut(SECTOR_COUNT_REG, (BYTE)(wSectorCount & 0xff));
      IDEOut(COMMAND_REG, COMMAND_READ_SECTORS);
#endif
    }

    for(i=0; i<wSectorCount; i++) {
      nError = WaitForInterrupt(DISKTIMEOUT);
      if (nError == IDE_OK) {
        bStatus = gbIntStatus;

        if (bStatus & STATUS_ERROR) {
          bErrorReg = IDEIn(ERROR_REG);
          nError    = IDE_ERROR;
          break;
        }

        for(x=0; x<IDE_SECTOR_SIZE; x=x+2) {
                
          pData[x] = pIDE[DATA_READ_REG_LOW];
          if (pDrive->bCompactFlash == TRUE) {
            pData[x+1] = pIDE[DATA_READ_REG_LOW];     
          } else {
            pData[x+1] = pIDE[DATA_READ_REG_HIGH];
          } 
          
        }
        pData += IDE_SECTOR_SIZE;

      } else {
        bStatus   = IDEIn(STATUS_REG);
        bErrorReg = IDEIn(ERROR_REG);
      }
    }

    bStatus = IDEIn(STATUS_REG);

  } /* SelectDevice */

  return (nError);
}

#if (IDE_SUPPORT_WRITE == 1)
/************************************************************/
/*  WriteSectors                                            */
/*                                                          */
/*  PIO data out command protocol, see ATAPI-4 r17 p227 9.8 */
/*                                                          */
/*  Return: IDE_OK, IDE_ERROR                               */
/************************************************************/
static BYTE WriteSectors (WORD   bDevice,      BYTE *pData, 
                          DWORD dwStartSector, WORD  wSectorCount)
{
  int     i, x;
  int     nError;
  BYTE    bValue;
  BYTE    bStatus;
  BYTE    bErrorReg;
  LPDRIVE pDrive;

#if (IDE_SUPPORT_CHS == 1)
  WORD    wCHSSector;
  WORD    wCHSHead;
  WORD    wCHSCylinder;
#endif

  pDrive = &sDrive[bDevice];
  nError = SelectDevice(bDevice);

#if (IDE_SUPPORT_CHS == 0)
  //
  // NO CHS support
  //
  if ((pDrive->wFlags & IDE_SUPPORT_LBA) == 0) {
    //
    // If we found a CHS device, this is an ERROR
    //
    nError = IDE_ERROR;
  }
#endif

  if (nError == IDE_OK) {

    if (pDrive->wFlags & IDE_SUPPORT_LBA) {
      //
      // LBA
      //
      bValue = (BYTE) ((bDevice << 4) | 0xE0 | (dwStartSector >> 24));
      IDEOut(DISK_HEAD_REG, bValue);
      IDEOut(LBA_16_23, (BYTE)(dwStartSector >> 16));
      IDEOut(LBA_8_15,  (BYTE)(dwStartSector >> 8));
      IDEOut(LBA_0_7,   (BYTE)(dwStartSector));

      IDEOut(SECTOR_COUNT_REG, (BYTE)(wSectorCount & 0xff));
      IDEOut(COMMAND_REG, COMMAND_WRITE_SECTORS);

#if (IDE_SUPPORT_CHS == 1)
    } else {
      //
      // CHS
      //
      wCHSSector    = (WORD) (dwStartSector % pDrive->wSectorsPerTrack) + 1;
      wCHSHead      = (WORD) (dwStartSector / pDrive->dwOneSide);
      dwStartSector = dwStartSector % pDrive->dwOneSide;
      wCHSCylinder  = (WORD) (dwStartSector / pDrive->wSectorsPerTrack);

      wCHSHead     |= (bDevice << 4) | 0xA0;

      IDEOut(DISK_HEAD_REG,     (BYTE)wCHSHead);
      IDEOut(SECTOR_REG,        (BYTE)wCHSSector);
      IDEOut(CYLINDER_LOW_REG,  (BYTE)(wCHSCylinder & 0x00FF));
      IDEOut(CYLINDER_HIGH_REG, (BYTE)(wCHSCylinder >> 8));

      IDEOut(SECTOR_COUNT_REG, (BYTE)(wSectorCount & 0xff));
      IDEOut(COMMAND_REG, COMMAND_WRITE_SECTORS);
#endif
    }

    if (WaitDRQ(CONTROLLERTIMEOUT) == IDE_OK) {

      bStatus = IDEIn(STATUS_REG);
      for(i=0; i<wSectorCount; i++) {

        for(x=0; x<IDE_SECTOR_SIZE; x=x+2) {
          if (pDrive->bCompactFlash == TRUE) {
            IDEOut(DATA_WRITE_REG_LOW,  pData[x]);
            IDEOut(DATA_WRITE_REG_LOW,  pData[x+1]);
          } else {
            IDEOut(DATA_WRITE_REG_HIGH, pData[x+1]);
            IDEOut(DATA_WRITE_REG_LOW,  pData[x]);
          }     
        }

        pData += IDE_SECTOR_SIZE;

        nError = WaitForInterrupt(DISKTIMEOUT);
        if (nError == IDE_OK) {
          bStatus = gbIntStatus;

          if ((bStatus & STATUS_ERROR) == STATUS_ERROR) {
            bErrorReg = IDEIn(ERROR_REG);
            nError    = IDE_ERROR;
            break;
          }
        } else {
          //
          // Error
          //
          nError = IDE_ERROR;
          break;
        }
      }

    } else {
      //
      // Device: Set error status (D)
      //
      WaitForInterrupt(DISKTIMEOUT);
      bErrorReg = IDEIn(ERROR_REG);
    } /* endif WaitDRQ */

  } /* SelectDevice */

  return (nError);
}
#endif

/*==========================================================*/
/*  DEFINE: All code exported                               */
/*==========================================================*/
/************************************************************/
/*  IDEInit                                                 */
/************************************************************/
int IDEInit (BYTE *pSectorBuffer)
{
  WORD i;
  int nError;
    
  //
  // Install IDE interrupt
  //
  nError = NutRegisterInterrupt(IRQ_INT7, IDEInterrupt, NULL);
  if (nError == FALSE) {
    EICRB |= 0xC0;   /* The rising edge between two samples */
                     /* of INT7 generates an interrupt      */
                     /* request.                            */
    sbi(EIMSK, INT7); 
  }
 
  
#if (USE_COMPACT_FLASH == FALSE)

  // 
  // I have some trouble with my IBM Travelstart DBCA-203240.
  // The disk does not work correct with 14.7456Mhz, 
  // therefore devide the clock by 2 and take some 
  // extra wait states
  //
  NutEnterCritical();

  XDIV  = 0xff;
  XMCRA = 0x02;

  NutExitCritical();

#endif

  
  for(i=0; i<IDE_MAX_SUPPORTED_DEVICE; i++) {
    memset((BYTE *)&sDrive[i], 0x00, sizeof(DRIVE));

    sDrive[i].bDevice       = (BYTE)i;
    sDrive[i].bCompactFlash = USE_COMPACT_FLASH;
  }

  //
  // Clear interrupt status
  //
  gbIntStatus = 0;

  pIDE = (unsigned char *)IDE_BASE_ADDRESS;
   
  ClearEvent();

  HardwareReset(&sDrive[0]);

#if (IDE_MAX_SUPPORTED_DEVICE == 2)
  IDEMountDevice (IDE_DRIVE_D, pSectorBuffer);
#endif

  IDEMountDevice (IDE_DRIVE_C, pSectorBuffer);

  return (nError);
}

/************************************************************/
/*  IDEMountDevice                                          */
/************************************************************/
int IDEMountDevice (BYTE bDevice, BYTE *pSectorBuffer)
{
  int     nError;
  LPDRIVE pDrive = NULL;

  IDELock();

  if (bDevice > IDE_MAX_SUPPORTED_DEVICE) {
    nError = IDE_ERROR;
  } else {
    pDrive = &sDrive[bDevice];

    pDrive->wFlags           = 0;
    
#if (IDE_SUPPORT_CHS == 1)
    //
    // CHS values
    //  
    pDrive->wCylinders       = 0;
    pDrive->wHeads           = 0;
    pDrive->wSectorsPerTrack = 0;
#endif  
    
    //
    // LBA value
    //
    pDrive->dwTotalSectors   = 0;

    nError = DeviceDiag(pDrive);
    if (nError != IDE_OK) {
      //
      // Try it again
      //
      nError = DeviceDiag(pDrive);
    }

    if (nError == IDE_OK) {
      nError = GetDeviceInfo(pDrive, pSectorBuffer);
      if (pDrive->wFlags & IDE_READY) {
        nError = IDE_OK;
      }
    }

  }

  IDEFree();

  return (nError);
}

/************************************************************/
/*  IDEGetTotalSectors                                      */
/************************************************************/
DWORD IDEGetTotalSectors (BYTE bDevice)
{
  DWORD   dwTotalSectors = 0;
  LPDRIVE  pDrive;

  IDELock();

  if (bDevice > IDE_MAX_SUPPORTED_DEVICE) {
    dwTotalSectors = 0;
  } else {
    pDrive          = &sDrive[bDevice];
    dwTotalSectors  = pDrive->dwTotalSectors;

    dwTotalSectors -= 64;
  }

  IDEFree();

  return (dwTotalSectors);
}

/************************************************************/
/*  IDEReadSectors                                          */
/************************************************************/
int IDEReadSectors (BYTE   bDevice,      void *pData, 
                    DWORD dwStartSector, WORD  wSectorCount)
{
  int      nError = IDE_OK;
  WORD     wReadCount;
  LPDRIVE  pDrive;
  BYTE    *pByte;

  IDELock();

  if (bDevice > IDE_MAX_SUPPORTED_DEVICE) {
    nError = IDE_DRIVE_NOT_FOUND;
  } else {
    pDrive = &sDrive[bDevice];
    if (pDrive->wFlags == 0) {
      nError = IDE_DRIVE_NOT_FOUND;
    } else {
      if( (dwStartSector+wSectorCount) > pDrive->dwTotalSectors) {
        nError = IDE_PARAM_ERROR;
      }
    }
  }

  if (nError == IDE_OK) {
    pByte = (BYTE *)pData;
    while (wSectorCount > 0) {
      
      if (wSectorCount < 256) {
        wReadCount = wSectorCount;
      } else {
        wReadCount = 256;
      }   

      nError = ReadSectors(bDevice, pByte, dwStartSector, wReadCount);
      if (nError != IDE_OK) {
        break;
      }

      dwStartSector += wReadCount;
      wSectorCount  -= wReadCount;
      pByte         += (wReadCount * IDE_SECTOR_SIZE);
    }
  }

  IDEFree();

  return (nError);
}

#if (IDE_SUPPORT_WRITE == 1)
/************************************************************/
/*  IDEWriteSectors                                         */
/************************************************************/
int IDEWriteSectors (BYTE   bDevice,      void *pData, 
                     DWORD dwStartSector, WORD  wSectorCount)
{
  int      nError = IDE_OK;
  WORD     wWriteCount;
  LPDRIVE  pDrive = NULL;
  BYTE    *pByte;

  IDELock();

  if (bDevice > IDE_MAX_SUPPORTED_DEVICE) {
    nError = IDE_DRIVE_NOT_FOUND;
  } else {
    pDrive = &sDrive[bDevice];
    if (pDrive->wFlags == 0) {
      nError = IDE_DRIVE_NOT_FOUND;
    } else {
      if( (dwStartSector+wSectorCount) > pDrive->dwTotalSectors) {
        nError = IDE_PARAM_ERROR;
      }
    }
  }

  if (pDrive->bCompactFlash == TRUE) {
    //
    // Sorry, I have not tested the write function
    // for the CompactFlash
    //
    nError = IDE_ERROR;
  }

  if (nError == IDE_OK) {
    pByte = (BYTE *)pData;
    while (wSectorCount > 0) {

      if (wSectorCount < 256) {
        wWriteCount = wSectorCount;
      } else {
        wWriteCount = 256;
      }   

      nError = WriteSectors(bDevice, pByte, dwStartSector, wWriteCount);
      if (nError != IDE_OK) {
        break;
      }

      dwStartSector += wWriteCount;
      wSectorCount  -= wWriteCount;
      pByte         += (wWriteCount * IDE_SECTOR_SIZE);
    }
  }

  IDEFree();

  return (nError);
}
#endif
